///////////////////////////////////////////////////////////////////////////
//
//   Copyright 2005,2006,2007,2008,2009 Alberto González Palomo
//   Author: Alberto González Palomo <http://www.matracas.org>
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
/////////////////////////////////////////////////////////////////////////////

// Notes:
// 2005-05-08 23:06: I've just found about a previous implementation of the
//   idea of a cascaded parser:
//    http://www.ai.sri.com/%7Eappelt/fastus-schabes.html
// 2006-11-12 11:24: There was a paper published in 1996 describing
//   cascade parsers: http://citeseer.ist.psu.edu/abney96partial.html

var parser_namespace = "http://www.matracas.org/ns/cascada";

function Ternary_tree()
{
  var self = this;

  function get_tree() { return tree; }

  function Entry(code_point)
  {
    this.code_point = code_point;
    //this.data = undefined;
    this.smaller = undefined;
    this.same = undefined;
    this.bigger = undefined;
  }

  function add_entry(pattern, data)
  {
    var entry;

    var match;
    match = search(pattern);

    var code_point;
    if (!tree)
      {
        code_point = pattern.charCodeAt(0);
        entry = new Entry(code_point);
        match.last_tested_entry  = entry;
        match.last_matched_entry = entry;
        match.length = 0;
        match.index = 1;
        tree = entry;
      }
    else if (match.last_matched_entry != match.last_tested_entry)
      {
        var parent;
        code_point = pattern.charCodeAt(match.index);
        entry = new Entry(code_point);
        parent = match.last_tested_entry;
        if      (code_point < parent.code_point) parent.smaller = entry;
        else if (code_point > parent.code_point) parent.bigger  = entry;
        ++match.index;
        match.last_tested_entry = entry;
      }

    while (match.index < pattern.length)
      {
        code_point = pattern.charCodeAt(match.index);
        entry = new Entry(code_point);
        match.last_tested_entry.same = entry;

        // From now on, we are matching as we go.
        ++match.index;
        match.last_tested_entry = entry;
      }

    match.length = match.index;

    // Now we have a complete match, so we store the associated data.
    // TODO: handle the case where this entry already had data. Should we
    //       at least issue a warning before overwriting the old data?
    //       Currently it's just overwritten.
    match.last_tested_entry.data = data;
  }

  function search(string, begin)
  {
    var match;
    match = {"length":0,
             "last_tested_entry":undefined, "last_matched_entry":undefined};

    if (!begin) begin = 0;

    match.index = begin;

    var node;
    node = tree;
    var string_length;
    string_length = string.length;
    var code_point;
    if (match.index < string_length) code_point = string.charCodeAt(match.index);
    while (node && match.index < string_length)
      {
        match.last_tested_entry = node;
        if (code_point == node.code_point)
          {
            ++match.index;
            code_point = string.charCodeAt(match.index);
            if ("data" in node)
              {
                // We take the longest (latest) match.
                // This could be extended easily to return an array with
                // all matches.
                match.length = match.index - begin;
                match.data = node.data;
              }
            node = node.same;
            match.last_matched_entry = match.last_tested_entry;
          }
        else if (code_point < node.code_point)
          {
            node = node.smaller;
          }
        else
          {
            node = node.bigger;
          }
      }

    return match;
  }

  // Constructor:
  var tree;

  // Public functions:
  this.get_tree = get_tree;
  this.Entry = Entry;
  this.add_entry = add_entry;
  this.search = search;
}

function Symbol(parser, replacement, identifier)
{
  var self = this;
  this.replacement = replacement;
  this.identifier = identifier;
}

function Term_class(parser, subclasses, replacement, identifier, complement)
{
  var self = this;

  function add(subclasses)
  {
    if (!subclasses) return;

    for (var c in subclasses)
      {
        if (!subclasses[c]) alert("Undefined tokens in class "
                                  + self.identifier + ":\n"
                                  + subclasses.toSource());
        if (parser.is_token(subclasses[c]))
          {
            self.subclasses.push(subclasses[c]);
            var term_subclass;
            term_subclass = parser.get_class(subclasses[c]);
            if (term_subclass) term_subclass.superclasses[identifier] = self;
          }
        else
          {
            self.subclasses.push(parser.symbol(subclasses[c]));
          }
      }

    recompile_dependent_rules();
  }

  function recompile_dependent_rules()
  {
    for (var r in self.rules) self.rules[r].compile();
    for (var s in self.superclasses)
      {
        self.superclasses[s].recompile_dependent_rules();
      }
  }

  // Constructor:
  this.replacement = replacement;
  this.identifier = identifier;
  this.complement = (complement?true:false);// Normalize to true/false.
  this.rules = {};// Rules that mention this class.
  this.superclasses = {};// Classes that contain this one as subclass.
  this.subclasses = [];
  add(subclasses);

  // Public functions:
  this.add = add;
  this.recompile_dependent_rules = recompile_dependent_rules;
}

function Rule(parser, unexpanded_pattern, replacement, identifier, context)
{
  var self = this;

  function expand_pattern(pattern, pattern_chars)
  {
    var token;
    var i;
    for (i in pattern)
      {
        token = pattern[i];
        var subclass;
        subclass = parser.get_class(token);
        if (subclass)
          {
            subclass.rules[replacement] = self;
            var subclass_tokens;
            subclass_tokens = [];
            parser.expand_subclasses(subclass, subclass_tokens);
            pattern_chars.push("[");
            if (subclass.complement) pattern_chars.push("^");
            pattern_chars.push(subclass_tokens.join(""));
            pattern_chars.push("]");
          }
        else
          {
            pattern_chars.push(token);
          }
      }
  }

  function compile()
  {
    var pattern_chars;
    pattern_chars = [];
    var expanded_pattern;
    var context_before;
    var context_after;

    pattern_chars.length = 0;// Redundant, just in case we reuse the array.
    expand_pattern(unexpanded_pattern, pattern_chars);
    expanded_pattern = pattern_chars.join("");

    if (context)
      {
        // Add the context patterns to "expanded_pattern".
        if (context[0] instanceof Array)
          {
            pattern_chars.length = 0;
            expand_pattern(context[0].join("|"), pattern_chars);
            context_before = pattern_chars.join("");
          }
        else
          {
            pattern_chars.length = 0;
            expand_pattern(context[0], pattern_chars);
            context_before = pattern_chars.join("");
          }
        if (context[1] instanceof Array)
          {
            pattern_chars.length = 0;
            expand_pattern(context[1].join("|"), pattern_chars);
            context_after = pattern_chars.join("");
          }
        else
          {
            pattern_chars.length = 0;
            expand_pattern(context[1], pattern_chars);
            context_after = pattern_chars.join("");
          }
        expanded_pattern
          = "(" + context_before + ")"
          + expanded_pattern
          + "(" + context_after + ")";
      }

    self.pattern = new RegExp(expanded_pattern, "g");
  }

  function match(string)         { return self.pattern.exec(string); }

  function last_index()          { return self.pattern.lastIndex; }

  function set_last_index(index) { self.pattern.lastIndex = index; }

  // Constructor:
  this.pattern = null;
  this.context = context;
  this.replacement = replacement;
  this.identifier = identifier;
  compile();

  // Public functions:
  this.compile = compile;
  this.match = match;
  this.last_index = last_index;
  this.set_last_index = set_last_index;
}

Rule.util =
  {
    any: function (pattern)
    {
      if (pattern.length > 1) return "(?:" + pattern + ")+";
      else                    return pattern + "+";
    },

    star: function (pattern)
    {
      if (pattern.length > 1) return "(?:" + pattern + ")*";
      else                    return pattern + "*";
    },

    or: function (patterns)
    {
      if (patterns instanceof Array) return "(?:" + patterns.join("|") + ")";
      else alert("Parser.or(patterns): patterns is not an array.");
      return patterns.toString();
    },

    opt: function (pattern)
    {
      if (pattern.length > 1) return "(?:" + pattern + ")?";
      else                    return pattern + "?";
    },

    not: function (patterns)
    {
      if (patterns instanceof Array) return "[^" + patterns.join("") + "]";
      else alert("Parser.not(patterns): patterns is not an array.");
      return patterns.toString();
    },

    literal: function (text)
    {
      return text.replace(/(\.|\?|\||\+|\*|\(|\)|\[|\]|\{|\}|\\|\^|\$)/g,
                          "\\$1");
    },

    input_begin: function ()
    {
      return "^";
    },

    input_end: function ()
    {
      return "$";
    }

  };

function Stage(model)
{
  if (model)
    {
      this.string = model.string;
      this.data = model.data.slice(0);
    }
  else
    {
      this.string = "";
      this.data  = [];
    }
}

function Parser()
{
  var self = this;

  function is_token(token)
  {
    return unicode_private_area_regexp.test(token);
  }

  function get_classes_with(token)
  {
    var classes;
    classes = [];
    var subclass;
    var term;
    term = get_term(token);

    for (var c in term_classes)
      {
        subclass = term_classes[c];
        for (var i in subclass.subclasses)
          if (token == subclass.subclasses[i])
            {
              classes.push(subclass.identifier);
              break;
            }
      }

    return classes;
  }

  function symbol_with_substring_comparison(identifier, is_regexp)
  {
    if (!identifier) throw "Parser.symbol(): no token given.";

    if (identifier in symbols) return symbols[identifier].replacement;

    ++last_term_index_assigned;
    var replacement;
    replacement = String.fromCharCode(last_term_index_assigned);
    var new_symbol;
    new_symbol = new Symbol(self, replacement, identifier);
    symbols[identifier] = new_symbol;
    terms_back[replacement] = new_symbol;

    if (identifier.length > longest_symbol_length)
      {
        longest_symbol_length = identifier.length;
      }

    token_table[identifier] = replacement;

    return replacement;
  }

  function symbol_with_ternary_tree(identifier, is_regexp)
  {
    if (!identifier) throw "Parser.symbol(): no token given.";

    if (identifier in symbols) return symbols[identifier].replacement;

    var match;
    match = ternary_tree.search(identifier);
    if (match.length == identifier.length) return match.data.replacement;

    ++last_term_index_assigned;
    var replacement;
    replacement = String.fromCharCode(last_term_index_assigned);
    var new_symbol;
    new_symbol = new Symbol(self, replacement, identifier);
    symbols[identifier] = new_symbol;
    ternary_tree.add_entry(identifier, new_symbol);
    terms_back[replacement] = new_symbol;

    token_table[identifier] = replacement;

    return replacement;
  }

  function symbol_with_rules(identifier, is_regexp)
  {
    var pattern;
    if (is_regexp) pattern = identifier;
    else           pattern = Rule.util.literal(identifier);

    var replacement;
    if (identifier in token_table) replacement = token_table[identifier];
    if (!replacement)
      {
        ++last_term_index_assigned;
        replacement = String.fromCharCode(last_term_index_assigned);
      }

    var new_rule;
    new_rule = new Rule(self, pattern, replacement, identifier, null);
    symbol_rules.push(new_rule);
    terms_back[replacement] = new_rule;

    token_table[identifier] = replacement;
    return replacement;
  }

  function term_class(identifier, subclasses)
  {
    return term_class_generic(identifier, subclasses, false);
  }

  function complement_term_class(identifier, subclasses)
  {
    return term_class_generic(identifier, subclasses, true);
  }

  function term_class_generic(identifier, subclasses, complement)
  {
    var replacement;
    if (identifier in token_table) replacement = token_table[identifier];
    if (!replacement)
      {
        ++last_term_index_assigned;
        replacement = String.fromCharCode(last_term_index_assigned);
      }
    var new_term_class;
    if (replacement in term_classes)
      {
        new_term_class = term_classes[replacement];
        if (new_term_class.complement != complement)
          {
            message("Error: new_term_class.complement != complement");
          }
        new_term_class.add(subclasses);
      }
    else
      {
        new_term_class = new Term_class(self,
                                        subclasses, replacement, identifier,
                                        complement);
        term_classes[replacement] = new_term_class;
      }

    token_table[identifier] = replacement;
    return replacement;
  }

  function rule(identifier, pattern, context)
  {
    var replacement;
    if (identifier in token_table) replacement = token_table[identifier];
    if (!replacement)
      {
        ++last_term_index_assigned;
        replacement = String.fromCharCode(last_term_index_assigned);
      }

    var new_rule;
    new_rule = new Rule(self, pattern, replacement, identifier, context);
    rules.push(new_rule);
    terms_back[replacement] = new_rule;

    token_table[identifier] = replacement;
    return replacement;
  }

  function action_rule(action, pattern, context)
  {
    var new_rule;
    new_rule = new Rule(self, pattern, action, action.name, context);
    rules.push(new_rule);

    return null;
  }

  function expand_subclasses(subclass, subclass_tokens)
  {
    var tokens;
    tokens = subclass.subclasses;
    subclass_tokens.push(tokens.join(""));
    var i;
    var token;
    for (i in tokens)
      {
        token = tokens[i];
        if (token in term_classes)
          {
            subclass = term_classes[token];
            if (subclass) expand_subclasses(subclass, subclass_tokens);
          }
      }
  }

  function current()
  {
    return String.fromCharCode(last_term_index_assigned + 1);
  }

  function reserve(identifier)
  {
    var replacement;
    replacement = String.fromCharCode(++last_term_index_assigned);
    if (identifier) token_table[identifier] = replacement;
    return replacement;
  }

  function get_term(token)
  {
    if (token in terms_back) return terms_back[token];
    else                     return null;
  }

  function get_class(token)
  {
    if (token in term_classes) return term_classes[token];
    else                       return null;
  }

  function parse(text)
  {
    var stage_stack;
    stage_stack = [];
    var last_stage;
    last_stage = tokenize(text);
    stage_stack.push(last_stage);
    var fence_match;
    do
      {
        last_stage = stage_stack[stage_stack.length - 1];
        fence_match = fence_regexp.exec(last_stage.string);
        if (fence_match && fence_match.index >= 0)
          {
            if (apply_rules(stage_stack, fence_match.index, fence_regexp.lastIndex))
              {
                fence_regexp.lastIndex = 0;
              }
          }
        else
          {
            apply_rules(stage_stack);
          }
      }
    while (fence_match && fence_match.index >= 0);

    return stage_stack;
  }

  function parse_to_XML(text)
  {
    var result;
    result = parse(text);
    //if ("application" in window) application.dump_properties(result[0]);
    var stage;
    stage = result[result.length - 1];
    var container_document_fragment;
    container_document_fragment = document.createDocumentFragment();
    var container;
    container = document.createElementNS(parser_namespace, "result");
    container_document_fragment.appendChild(container);

    for (var i in stage.string)
      {
        append_XML_node(container, stage.data[i], 1);
      }

    return container_document_fragment;
  }

  function append_XML_node(parent, data, level)
  {
    if (!data || level > 100)
      {
        parent.appendChild(document.createTextNode("[NO DATA]"));
        return;
      }

    var max_level;
    max_level = 8;
    var node;
    if (data.stage_index > 0)
      {
        node = document.createElementNS(parser_namespace, data.identifier);
      }
    else
      {
        if (data.identifier)
          {
            node = document.createElementNS(parser_namespace, "token");
            node.setAttribute("literal", data.identifier);
          }
        else
          {
            node = document.createElementNS(parser_namespace, "text");
            node.setAttribute("literal", data.token);
          }
      }
    if ("children" in data)
      {
        for (var i in data.children)
          {
            append_XML_node(node, data.children[i], level + 1);
          }
        node.setAttribute("begin", node.firstChild.getAttribute("begin"));
        node.setAttribute("end",   node.lastChild.getAttribute("end"));
      }
    else
      {
	node.setAttribute("begin", data.begin);
	node.setAttribute("end",   data.end);
      }
    parent.appendChild(node);
  }

  function unparse_from_XML(parse_tree)
  {
    // The parse tree can be either a document fragment (as produced by
    // parse_to_XML), or a DOM node.
    var linear_content;
    linear_content = "";
    var tokens;
    tokens = [];
    collect_tokens_from_parse_tree(parse_tree, tokens);
    if (tokens.length)
      {
        var tentative_linear_content;
        tentative_linear_content = tokens.join("");
        var recognized_tokens;
        recognized_tokens = tokenize_as_strings(tentative_linear_content);
	//alert("tokens.length: " + tokens.length + ", recognized_tokens.length:" + recognized_tokens.length + "\n" + tokens.toSource() + "\n" + recognized_tokens.toSource());
        var i;
	var last_correct_token;
	last_correct_token = -1;
	linear_content = "";
	while ((last_correct_token + 1) < tokens.length)
	  {
	    for (i = last_correct_token + 1; i < recognized_tokens.length; ++i)
	      {
		linear_content += tokens[i];
		if (recognized_tokens[i] != tokens[i])
		  {
		    linear_content += " ";
		    tentative_linear_content = linear_content + tokens.slice(i+1).join("");
		    recognized_tokens = tokenize_as_strings(tentative_linear_content);
		    break;
		  }
	      }
	    last_correct_token = i;
	  }
        if (tokens.length != recognized_tokens.length)
          {
            var powerful_message;
            powerful_message = "\nafter '";
	    for (i = 0; i < tokens.length && i < recognized_tokens.length; ++i)
	      if (tokens[i] == recognized_tokens[i])
		{
		  powerful_message += tokens[i];
		}
	      else
		{
		  powerful_message += "'\n" + i + " [" + tokens[i] + "] != [" + recognized_tokens[i].token + "]\n";
		  break;
		}
            message("Error: the given parse tree contains tokens that would not be recognized by the parser." + powerful_message);
          }
      }

    return linear_content;
  }
  function collect_tokens_from_parse_tree(parse_tree, tokens)
  {
    var child;
    child = parse_tree.firstChild;
    while (child)
      {
        switch (child.localName)
          {
          case "token":
          case "text":
            tokens.push(child.getAttribute("literal"));
            break;
          default:
            collect_tokens_from_parse_tree(child, tokens);
            break;
          }

        child = child.nextSibling;
      }
  }
  function tokenize_as_strings(string)
  {
    var s = [];
    var tokens;
    tokens = tokenize(string).data;

    for (var i = 0; i < tokens.length; ++i)
      {
	s.push(string.substring(tokens[i].begin, tokens[i].end));
      }

    return s;
  }

  // Returns whether some rule matched or not.
  // Begin and end are optional, and ignored unless both are distinct from 0.
  function apply_rules(stage_stack, begin, end)
  {
    var at_least_one_rule_applied;
    at_least_one_rule_applied = false;

    var stage;
    var stage_index;
    stage = new Stage();
    stage_index = stage_stack.length;
    var last_stage;
    var string;
    last_stage = stage_stack[stage_stack.length - 1];
    string = last_stage.string;
    var initial_string_length;
    initial_string_length = string.length;

    if (!begin) begin = 0;
    if (!end)   end   = 0;
    var rules_applied_in_iteration;
    do
      {
        var match_begin;
        var match_end;
        var substring_to_match;
        if (begin >= 0 && end > begin)
          {
            match_begin        = begin;
            match_end          = end - (initial_string_length - string.length);
            substring_to_match = string.substring(0, match_end);
          }
        else
          {
            match_begin        = 0;
            match_end          = string.length;
            substring_to_match = string;
          }

        var r;
        for (r = 0; r < rules.length; ++r)
          {
            rules[r].set_last_index(match_begin);
          }
        rules_applied_in_iteration = 0;
        for (r = 0; r < rules.length; ++r)
          {
            var rule;
            var data;

            var match;
            var p;

            // The first version of the rest of this block was the following
            // line, but we need to keep track of how many tokens are matched
            // by each rule, not just replace them.
            //stage = stage.replace(rules[r].pattern, rules[r].replacement);
            rule = rules[r];

            if ((match = rule.match(substring_to_match)))
              {
                if (rule.context)
                  {
                    var last_group_index;
                    last_group_index = match.length - 1;
                    if (!match[1] && !match[last_group_index])
                      {
                        message("Warning: there is a null context match in " + rule.identifier + ":\n"
                                + match.toSource()
                                );
                      }
                    // Check the context, and discard the match if it's not
                    // there.
                    if (match[1]) match.index += match[1].length;
                    if (match[last_group_index]) rule.set_last_index(rule.last_index() - match[last_group_index].length);
                    // The following shouldn't happen with sane rules, but
                    // this way we avoid the program from blocking in such
                    // a happenstance.
                    if (match.index == rule.last_index())
                      {
                        alert("The rule " + rule.identifier +
                              " has a context match, but its own match is void:\n"
                              + rule);
                        rule.set_last_index(match.index + 1);
                        continue;// for (r in rules)
                      }
                  }

                var replacement;
                replacement = rule.replacement;
                if ("function" == typeof(replacement))
                  {
                    message("Matched function rule " + rule.identifier);
                    // If the action function returns false, it means that
                    // there should not be any replacement of the matched
                    // input. For instance, when declaring variables.
                    replacement = rule.replacement(self, rule, substring_to_match.substring(match.index, rule.last_index()));
                    if (!replacement)
                      {
                        r = 0;
                        continue;
                      }
                  }

                // We only count rules that cause changes in the stage.
                at_least_one_rule_applied = true;
                ++rules_applied_in_iteration;

                data = {};
                data.token = replacement;
                data.identifier = rule.identifier;
                data.begin = match.index;
                data.end   = rule.last_index();
                data.stage_index = stage_index;
                data.children = [];
                var is_mergeable;
                is_mergeable = self.merge_stages;
                var last_stage_data;
                for (p = data.begin; p < data.end; ++p)
                  {
                    last_stage_data = last_stage.data[p];
                    data.children.push(last_stage_data);
                    is_mergeable &= (last_stage_data.stage_index + 1
                                     < stage_index);
                  }
                var replacement_count;
                replacement_count = data.end - data.begin;
                if (is_mergeable)
                  {
                    // Assuming replacement.length = 1:
                    last_stage.data.splice(data.begin,
                                           replacement_count, data);

                    last_stage.string
                      = string.substring(0, data.begin)
                      + data.token
                      + string.substring(data.end);
                    string = last_stage.string;
                  }
                else
                  {
                    // Assuming replacement.length = 1:
                    stage.data  = last_stage.data.slice(0);// Shallow copy.
                    stage.data.splice(data.begin, replacement_count, data);

                    stage.string
                      = string.substring(0, data.begin)
                      + data.token
                      + string.substring(data.end);

                    stage_stack.push(stage);

                    stage = new Stage();
                    stage_index = stage_stack.length;
                    last_stage = stage_stack[stage_stack.length - 1];
                    string = last_stage.string;
                  }

                break;// for (r in rules): Apply only one rule per iteration.
              }
          }
      }
    while (rules_applied_in_iteration > 0);

    return at_least_one_rule_applied;
  }

  function tokenize_with_ternary_tree(text)
  {
    var tokens = new Stage();
    var cursor = 0;
    var literal_mode = false;
    var test_string;
    var test_length;

    var symbol;
    var data;
    while (cursor < text.length)
      {
        symbol = null;
	if (!literal_mode || '"' == text[cursor])
	  {
        var match;
        match = ternary_tree.search(text, cursor);
        if ("data" in match) symbol = match.data;
        if (implicit_variable_name_regexp)
          {
            test_string = text.substr(cursor);
            var variable_name_match;
            variable_name_match = implicit_variable_name_regexp.exec(test_string);
            if (variable_name_match
                && variable_name_match[0].length > match.length)
              {
                match = variable_name_match[0];
                test_string = test_string.substr(0, match.length);
                self.term_class("object_variable", self.symbol(test_string));
                symbol = symbols[test_string];
              }
          }
	    if ('"' == text[cursor]) literal_mode = !literal_mode;
	  }
        if (symbol)
          {
            // The actual match can be shorter than the test length if
            // there are not so many characters left in "text". Therefore,
            // we correct the test_length variable to reflect the actual
            // length matched.
            test_length = match.length;
            data = {};
            data.token = symbol.replacement;
            data.identifier = symbol.identifier;
            data.begin = cursor;
            data.end   = cursor + match.length;
            data.stage_index = 0;
            tokens.data.push(data);
            tokens.string += data.token;
            cursor += test_length;
          }
        else
          {
            if (self.ignore_space && text[cursor].match(/\s/))
              {
                cursor += 1;
              }
            else
              {
                data = {};
                data.token = text[cursor];
                data.identifier = null;
                data.begin = cursor;
                data.end   = cursor + 1;
                data.stage_index = 0;
                tokens.data.push(data);
                tokens.string += data.token;
                cursor += 1;
              }
          }
      }

    return tokens;
  }

  function tokenize_with_substring_comparison(text)
  {
    var tokens;
    tokens = new Stage();
    var cursor;
    cursor = 0;
    var test_string;
    var test_length;

    var symbol;
    var data;
    while (cursor < text.length)
      {
        test_length = longest_symbol_length;
        if (cursor+test_length > text.length) test_length = text.length-cursor;
        symbol = null;
        while (!symbol && test_length > 0)
          {
            test_string = text.substr(cursor, test_length--);
            symbol = symbols[test_string];
            if (!symbol && implicit_variable_name_regexp)
              {
                var variable_name_match;
                variable_name_match = implicit_variable_name_regexp.exec(test_string);
                if (variable_name_match && variable_name_match[0].length == test_length)
                  {
                    // This is wasteful as it creates symbols for each
                    // intermediate string made while typing, but it does so
                    // only if they are not already defined, so in practice
                    // it's not so bad. It would be nice though to optimize it.
                    self.term_class("object_variable", self.symbol(test_string));
                    symbol = symbols[test_string];
                  }
              }
          }
        if (symbol)
          {
            // The actual match can be shorter than the test length if
            // there are not so many characters left in "text". Therefore,
            // we correct the test_length variable to reflect the actual
            // length matched.
            test_length = test_string.length;
            data = {};
            data.token = symbol.replacement;
            data.identifier = symbol.identifier;
            data.begin = cursor;
            data.end   = cursor + test_length;
            data.stage_index = 0;
            tokens.data.push(data);
            tokens.string += symbol.replacement;
            cursor += test_length;
          }
        else
          {
            if (self.ignore_space && text[cursor].match(/\s/))
              {
                cursor += 1;
              }
            else
              {
                data = {};
                data.token = text[cursor];
                data.identifier = null;
                data.begin = cursor;
                data.end   = cursor + 1;
                data.stage_index = 0;
                tokens.data.push(data);
                tokens.string += text[cursor];
                cursor += 1;
              }
          }
      }

    return tokens;
  }

  // Constructor:
  // Implementation switches, until we can settle on the ternary tree one.
  //var implementation = "substring_comparison";// 122 ms
  var implementation = "ternary_tree";// 120 ms
  var symbol;
  var tokenize;
  switch (implementation)
    {
    case "substring_comparison":
      symbol = symbol_with_substring_comparison;
      tokenize = tokenize_with_substring_comparison;
      // apply_rules
      break;
    case "ternary_tree":
      symbol = symbol_with_ternary_tree;
      tokenize = tokenize_with_ternary_tree;
      // apply_rules
      break;
    case "symbols_as_rules":
      symbol = symbol_with_rules;
      tokenize = tokenize_with_rules;
      // apply_rules
      break;
    default:
      message("Unknown implementation: " + implementation);
      symbol = symbol_with_substring_comparison;
      tokenize = tokenize_with_strings;
      // apply_rules
    }

  var implicit_variable_name_regexp = null;
  if ("watch" in this)
    {
      // this.watch is not defined in KHTML, and this code is currently not
      // needed anyway.
      this.watch("implicit_variable_name_pattern", function (property, oldValue, newValue)
                 {
                   try
                     {
                       if (newValue) implicit_variable_name_regexp = new RegExp("^" + newValue);
                       else          implicit_variable_name_regexp = null;
                     }
                   catch (exception)
                     {
                       message("Error: implicit_variable_name_pattern: " + exception);
                       newValue = oldValue;
                     }

                   return newValue;
                 }
                 );
    }

  var token_table = {};
  this.token_table = token_table;

  // Only needed if tokenizing with substrings.
  var symbols = {};
  var longest_symbol_length = 0;

  // Only needed if tokenizing with the ternary tree.
  var ternary_tree = new Ternary_tree();

  // Only needed if tokenizing with rules.
  var symbol_rules = [];

  var rules = [];
  var terms_back = {};
  var term_classes = {};
  var unicode_private_area_begin = 0xE000;
  var unicode_private_area_end   = 0xF8FF;
  var unicode_private_area_regexp = /[\uE000-\uF8FF]/;
  var last_term_index_assigned = unicode_private_area_begin - 1;

  // TODO: make the fence characters configurable
  var fence_regexp;
  fence_regexp = new RegExp
    (symbol("(")+"[^"+symbol("(")+symbol(")")+"]+"+symbol(")") + "|" +
     symbol("[")+"[^"+symbol("[")+symbol("]")+"]+"+symbol("]") + "|" +
     symbol("{")+"[^"+symbol("{")+symbol("}")+"]+"+symbol("}") + "|" +
     "^"        +"[^"+symbol("(")+symbol(")")+"]+"+symbol(")") + "|" +
     "^"        +"[^"+symbol("[")+symbol("]")+"]+"+symbol("]") + "|" +
     "^"        +"[^"+symbol("{")+symbol("}")+"]+"+symbol("}") + "|" +
     symbol("(")+"[^"+symbol("(")+symbol(")")+"]+"+        "$" + "|" +
     symbol("[")+"[^"+symbol("[")+symbol("]")+"]+"+        "$" + "|" +
     symbol("{")+"[^"+symbol("{")+symbol("}")+"]+"+        "$",
     "g"// Important! Otherwise fence_regexp.lastIndex is not set.
     );// /\([^\)]*\)|\[[^\]]*\]|\{[^\}]*\}/g;
  //message("Fence regexp = " + fence_regexp.toString().replace(symbol("("), "\\(", "g").replace(symbol(")"), "\\)", "g").replace(symbol("["), "\\[", "g").replace(symbol("]"), "\\]", "g").replace(symbol("{"), "\\{", "g").replace(symbol("}"), "\\}", "g"));

  // Public functions:
  this.is_token    = is_token;
  this.get_classes_with = get_classes_with;
  this.symbol      = symbol;
  this.term_class  = term_class;
  this.complement_term_class = complement_term_class;
  this.rule        = rule;
  this.action_rule = action_rule;
  this.any         = Rule.util.any;
  this["+"]        = Rule.util.any;
  this.star        = Rule.util.star;
  this.opt_any     = Rule.util.star;
  this["*"]        = Rule.util.star;
  this.or          = Rule.util.or;
  this["|"]        = Rule.util.or;
  this.opt         = Rule.util.opt;
  this["?"]        = Rule.util.opt;
  this.not         = Rule.util.not;
  this.input_begin = Rule.util.input_begin;
  this.input_end   = Rule.util.input_end;
  this.current = current;
  this.reserve = reserve;
  this.get_term = get_term;
  this.get_class = get_class;
  this.expand_subclasses = expand_subclasses;
  this.parse = parse;
  this.parse_to_XML = parse_to_XML;
  this.unparse_from_XML = unparse_from_XML;
  // Public properties:
  this.ignore_space = false;
  this.merge_stages = true;
  this.implicit_variable_name_pattern = null;
}
